#ifdef WIN32
#pragma comment(lib, "c:\\TopfieldDisk\\DataFiles\\SDK\\TAP_SDK_DLL")
#include "TAP_Emulations.h"
#endif

#include "tap.h"
//#include "exTap.h"
#include "menu.h"
#include "dlmalloc.h"
#include "rasterfontbrowser.h"

#define MULTIPLIER 25173
#define INCREMENT 13849
#define MODULUS 65536

void init_genrand(unsigned long s);
long genrand_int31(void);
void ExtractExtension(char *name, char *extension);
void logText(const char* text, bool close_log);
void ExtractDirectory(char *path, char *significant_part);
void ChangeToRoot(char *new_path);
void ChangeDirRootFromUnknown();

extern dword random_seed;
extern char log_buffer[512];
extern char tracked_dir[256];

static word* sysID = (word*)0xa3fffffa;
static char tmp[100];
static char tmp2[100];
static char tmp3[256];
extern char buffer4[256];
extern char buffer5[256];

int getWidthRounded(char *text, byte font_size, bool rounded)
{
	int width = 0, length = 0;

	if (strcmp(font_name, "Default") != 0) {
		switch (font_size) {
			case FNT_Size_1419:	
				if (fonts[current_font].small_font)
					CalcSizeRounded(fonts[current_font].small_font, text, &width, &length, rounded);
				else
					return TAP_Osd_GetW(text, 0, font_size);
				break;
			case FNT_Size_1622:	
				if (fonts[current_font].medium_font)
					CalcSizeRounded(fonts[current_font].medium_font, text, &width, &length, rounded);
				else
					return TAP_Osd_GetW(text, 0, font_size);
				break;
			case FNT_Size_1926:	
				if (fonts[current_font].large_font) {
					CalcSizeRounded(fonts[current_font].large_font, text, &width, &length, rounded);
				}
				else
					return TAP_Osd_GetW(text, 0, font_size);
				break;
		}
	}
	else
		return TAP_Osd_GetW(text, 0, font_size);

	return width;
}

int getWidth(char *text, byte font_size)
{
	return getWidthRounded(text, font_size, TRUE);
}

word ChangeDir(char *path)
{
	word success;
#ifndef WIN32
	extern int _appl_version;
#else
	int _appl_version = 0x1000;
#endif

	// do the change
	//success = (_appl_version >= 0x1200) ? TAP_Hdd_ChangeDir(path) : !TAP_Hdd_ChangeDir(path);
#ifndef WIN32
	success = TAP_Hdd_ChangeDir(path);
#else
	success = !TAP_Hdd_ChangeDir(path);
#endif
	//if (write_log) TAP_SPrint(log_buffer, "Changed dir %s (%s)", path, success ? "success" : "failure");
	//if (write_log) logText(log_buffer, FALSE);

	// update the current path
	if (success || strcmp(path, "ProgramFiles") == 0 || strcmp(path, "DataFiles") == 0 || strcmp(path, "MP3") == 0) {
		if (strcmp(path, "..") == 0) {
			strcpy(log_buffer, tracked_dir);
			ExtractDirectory(log_buffer, tracked_dir);
		}
		else {
			if (strcmp(tracked_dir, "/") != 0)
				strcat(tracked_dir, "/");
			strcat(tracked_dir, path);
		}
	}

	return success;
}

word ChangeDirWithoutLogging(char *path)
{
	word success;
#ifndef WIN32
	extern int _appl_version;
#else
	int _appl_version = 0x1000;
#endif

	// do the change
	//success = (_appl_version >= 0x1200) ? TAP_Hdd_ChangeDir(path) : !TAP_Hdd_ChangeDir(path);
	success = TAP_Hdd_ChangeDir(path);

	// update the current path
	if (success || strcmp(path, "ProgramFiles") == 0 || strcmp(path, "DataFiles") == 0 || strcmp(path, "MP3") == 0) {
		if (strcmp(path, "..") == 0) {
			strcpy(log_buffer, tracked_dir);
			ExtractDirectory(log_buffer, tracked_dir);
		}
		else {
			if (strcmp(tracked_dir, "/") != 0)
				strcat(tracked_dir, "/");
			strcat(tracked_dir, path);
		}
	}

	return success;
}

bool isMp3Directory(char *file_name)
{
	return (strcmp(file_name, "MP3") == 0);
}

bool isParentDirectory(char *file_name)
{
	return (strcmp(file_name, "..") == 0);
}

void getDrawDimensionsAndOffset(int bitmap_width, int bitmap_height, int *calc_width, int *calc_height, int *calc_offset_x, int *calc_offset_y, 
								int widescreen_factor, bool cache, bool thumbnail, bool rotated_sideways, bool full_size, bool album_art, bool compact_art)
{
	int w, h, width = bitmap_width, height = bitmap_height;
	int xl, yl, max_width, max_height;
	int ratio, percentage;
	int actual_offset_x = 0, actual_offset_y = 0;

	// set up the ideal width and height
	if (compact_art) {
		xl = max_width = 200; yl = max_height = 200;
	}
	ef (thumbnail) {
		xl = max_width = 150; yl = max_height = 112;
	}
	ef (cache || full_size) {
		xl = max_width = 717; yl = max_height = 539;
	}
	else {
		xl = max_width = 600; yl = max_height = 450;
	}

	// first go at scaling down the dimensions to fit OSD
	if (width > height) {
		if (width > xl) {
			yl = height*xl/width;
		}
		ef (height > yl)
		{
			xl = width * yl / height;
			yl = height * yl / height;
		} else {
			xl = width;
			yl = height;
		}
	} else {
		if (height > yl) {
			xl = width*yl/height;
		} else {
			xl = width;
			yl = height;
		}
	}

	// if height still too large, then scale to fit
	if (rotated_sideways) {
		if (yl > max_width)	{
			xl = xl * max_width / yl;
			yl = max_width;
		}

		if (xl > max_height) {
			yl = yl * max_height / xl;
			xl = max_height;
		}
	}
	else {
		if (yl > max_height) {
			xl = xl * max_height / yl;
			yl = yl * max_height / yl;
		}
	}

	// if the photo is smaller that standard viewer size, then zoom it
	if ((album_art || full_size) && !cache) {
		if (rotated_sideways && (xl < 600 && yl < 450)) {
			if (yl > xl) {
				if (yl * 2 > max_width) {
					xl = xl * max_width / yl;
					yl = max_width;
				}
				else {
					yl *= 2;
					xl *= 2;
				}
			}
			else {
				if (xl * 2 > max_height) {
					yl = yl * max_height / xl;
					xl = max_height;
				}
				else {
					yl *= 2;
					xl *= 2;
				}
			}
		}
		ef (!rotated_sideways && (xl < 600 && yl < 450)) {
			if (xl > yl) {
				if (xl * 2 > max_width) {
					yl = yl * max_width / xl;
					xl = max_width;
				}
				else {
					xl *= 2;
					yl *= 2;
				}
			}
			else {
				if (yl * 2 > max_height) {
					xl = xl * max_height / yl;
					yl = max_height;
				}
				else {
					xl *= 2;
					yl *= 2;
				}
			}
		}

		// if height still too large, then scale to fit
		if (rotated_sideways) {
			if (yl > max_width)	{
				xl = xl * max_width / yl;
				yl = max_width;
			}

			if (xl > max_height) {
				yl = yl * max_height / xl;
				xl = max_height;
			}
		}
		else {
			if (yl > max_height) {
				xl = xl * max_height / yl;
				yl = yl * max_height / yl;
			}
		}
	}

	// allow for width to be reduced by widescreen scaling
	if (!cache && !compact_art && widescreen_factor > 0)
		xl = (xl * 100) / (widescreen_factor + 100);
	
	// calculate offsets
	if (!cache && !compact_art) {
		if (full_size) {
			if (rotated_sideways) {
				actual_offset_x = yl < max_width ? ((max_width - yl) / 2) : 0;
				actual_offset_y = xl < max_height ? ((max_height - xl) / 2) : 0;		
			}
			else {
				actual_offset_x = xl < max_width ? ((max_width - xl) / 2) : 0;
				actual_offset_y = yl < max_height ? ((max_height - yl) / 2) : 0;
			}
		}
		else {
			if (rotated_sideways) {
				actual_offset_x = yl < max_width ? ((max_width - yl) / 2) : 0;
				actual_offset_y = xl < max_height ? ((max_height - xl) / 2) : 0;
			}
			else {
				actual_offset_x = xl < max_width ? ((max_width - xl) / 2) : 0;
				actual_offset_y = yl < max_height ? ((max_height - yl) / 2) : 0;
			}
		}

		// if image is higher than it is taller, then will need to shrink horizontally, so adjust offset accordingly
		w = width; h = height;
		if (widescreen_factor > 0)
			w = width * 100 / (widescreen_factor + 100);
		if (w > max_width && h > max_height && w < h)
		{
			ratio = (w * 100) / height;
			percentage = (ratio * 100) / 133;
			actual_offset_x = ((max_width - ((max_width * percentage) / 100)) / 2);
		}
	}

	// finally, set the needed variables
	*calc_width = xl;
	*calc_height = yl;
	if (!cache && !compact_art) {
		*calc_offset_x = actual_offset_x;
		*calc_offset_y = actual_offset_y;
	}
}

void seedRandom()
{
	init_genrand(TAP_GetTick());
}

dword getRandom()
{
	return genrand_int32();
}

bool InProgramFilesFolder(char* path)
{
    if (strlen(path) > 12 && strncmp(path, "/ProgramFiles", 13) == 0) 
		return TRUE;

	return FALSE;
}

bool InFolder(char* trial, char* folder)
{
    if (strlen(trial) >= strlen(folder) && strncmp(trial, folder, strlen(folder)) == 0) 
		return TRUE;

	return FALSE;
}

bool InMp3FilesFolder(char* path)
{
    if (strlen(path) > 3 && (strncmp(path, "/MP3", 4) == 0)) 
		return TRUE;

	return FALSE;
}

bool IsDataFilesFolder(char* path)
{
	if (strcmp(path, "/DataFiles") == 0) 
		return TRUE;

	return FALSE;
}

bool InDataFilesFolder(char* path)
{
    if (strlen(path) > 9 && (strncmp(path, "/DataFiles", 10) == 0)) 
		return TRUE;

	return FALSE;
}

bool isLog(char *name)
{
    static char tmp[128];

	if (strlen(name) < 5)
		return FALSE;

    strcpy(tmp, name);
    strlwr(tmp);

    if (strcmp(tmp + strlen(tmp) - 4, ".log" ) == 0
		|| strcmp(tmp + strlen(tmp) - 4, ".ini" ) == 0
		|| strcmp(tmp + strlen(tmp) - 4, ".mcf" ) == 0
		|| strcmp(tmp + strlen(tmp) - 4, ".log" ) == 0
		|| strcmp(tmp + strlen(tmp) - 4, ".cfg" ) == 0
		|| strcmp(tmp + strlen(tmp) - 4, ".mei" ) == 0
		|| strcmp(tmp + strlen(tmp) - 4, ".txt" ) == 0
		|| strcmp(tmp + strlen(tmp) - 4, ".tgd" ) == 0
		|| strcmp(tmp + strlen(tmp) - 5, ".html" ) == 0
		|| strcmp(tmp + strlen(tmp) - 4, ".htm" ) == 0
		|| strcmp(tmp + strlen(tmp) - 4, ".xml" ) == 0
		|| strcmp(tmp + strlen(tmp) - 4, ".skn" ) == 0
		|| strcmp(tmp + strlen(tmp) - 2, ".c" ) == 0
		|| strcmp(tmp + strlen(tmp) - 2, ".h" ) == 0
		|| strcmp(tmp + strlen(tmp) - 4, ".doc" ) == 0
		|| strcmp(tmp + strlen(tmp) - 4, ".xls" ) == 0
		|| strcmp(tmp + strlen(tmp) - 4, ".dat" ) == 0
		|| strcmp(tmp + strlen(tmp) - 4, ".pdf" ) == 0
		|| strcmp(tmp + strlen(tmp) - 4, ".bin" ) == 0
		|| strcmp(tmp + strlen(tmp) - 4, ".chn" ) == 0)
        return TRUE;

    return FALSE;
}

bool isMp3(char *name)
{
    strcpy(tmp, name);
    strlwr(tmp);

	if (strlen(tmp) < 5) return FALSE;

    if (strcmp(tmp + strlen(tmp) - 4, ".mp3" ) == 0)
        return TRUE;

    return FALSE;
}

bool isJpg(char *name)
{
    strcpy(tmp, name);
    strlwr(tmp);

	if (strlen(tmp) < 5) return FALSE;

    if (strcmp(tmp + strlen(tmp) - 4, ".jpg" ) == 0)
        return TRUE;

	if (strlen(tmp) < 6) return FALSE;

    if (strcmp(tmp + strlen(tmp) - 5, ".jpeg" ) == 0)
        return TRUE;

    return FALSE;
}

bool isBmp(char *name)
{
    strcpy(tmp, name);
    strlwr(tmp);

	if (strlen(tmp) < 5) return FALSE;

    if (strcmp(tmp + strlen(tmp) - 4, ".bmp" ) == 0)
        return TRUE;

    return FALSE;
}

bool isGif(char *name)
{
    strcpy(tmp, name);
    strlwr(tmp);

	if (strlen(tmp) < 5) return FALSE;

    if (strcmp(tmp + strlen(tmp) - 4, ".gif" ) == 0)
        return TRUE;

    return FALSE;
}

bool isPhoto(char *name)
{
    return isJpg(name) || isBmp(name) || isGif(name);
}

bool isRec(char *name)
{
    strcpy(tmp, name);
    strlwr(tmp);

	if (strlen(tmp) < 5) return FALSE;

    if (strcmp(tmp + strlen(tmp) - 4, ".rec" ) == 0)
        return TRUE;

    return FALSE;
}

bool isSkin(char *name)
{
    strcpy(tmp, name);
    strlwr(tmp);

	if (strlen(tmp) < 5) return FALSE;

    if (strcmp(tmp + strlen(tmp) - 4, ".mcf" ) == 0)
        return TRUE;

    return FALSE;
}

bool isFont(char *name)
{
    strcpy(tmp, name);
    strlwr(tmp);

	if (strlen(tmp) < 5) return FALSE;

    if (strcmp(tmp + strlen(tmp) - 4, ".fnt" ) == 0)
        return TRUE;

    return FALSE;
}

bool isTAP(char *name)
{
    strcpy(tmp, name);
    strlwr(tmp);

	if (strlen(tmp) < 5) return FALSE;

    if (strcmp(tmp + strlen(tmp) - 4, ".tap" ) == 0)
        return TRUE;

    return FALSE;
}

// true if extension is in format .9 or .99
bool isScene(char *name)
{
	word extension_length = 0;
	char extension[10];

	ExtractExtension(name, extension);
	extension_length = strlen(extension);
	if (extension_length < 2 || extension_length > 3)
		return FALSE;

	// check between 1-9
	if (extension_length == 2)
		if (extension[1] >= 49 && extension[1] <= 57)
			return TRUE;

	// check between 1-9, then 0-9
	if (extension[1] >= 49 && extension[1] <= 57
			&& extension[2] >= 48 && extension[2] <= 57)
		return TRUE;

	return FALSE;
}

bool isRecycled(char *name)
{
    strcpy(tmp, name);
    strlwr(tmp);

	if (strlen(tmp) < 5) return FALSE;

    if (strcmp(tmp + strlen(tmp) - 4, ".del" ) == 0)
        return TRUE;

    return FALSE;
}

bool isRecycledJpg(char *name)
{
    strcpy(tmp, name);
    strlwr(tmp);

	if (strlen(tmp) < 9) return FALSE;

    if (strcmp(tmp + strlen(tmp) - 8, ".jpg.del" ) == 0)
        return TRUE;

    return FALSE;
}

bool isRecycledBmp(char *name)
{
    strcpy(tmp, name);
    strlwr(tmp);

	if (strlen(tmp) < 9) return FALSE;

    if (strcmp(tmp + strlen(tmp) - 8, ".bmp.del" ) == 0)
        return TRUE;

    return FALSE;
}

bool isRecycledGif(char *name)
{
    strcpy(tmp, name);
    strlwr(tmp);

	if (strlen(tmp) < 9) return FALSE;

    if (strcmp(tmp + strlen(tmp) - 8, ".gif.del" ) == 0)
        return TRUE;

    return FALSE;
}

bool isRecycledPhoto(char *name)
{
    return isRecycledJpg(name) || isRecycledBmp(name) || isRecycledGif(name);
}

bool isRecycledMp3(char *name)
{
    strcpy(tmp, name);
    strlwr(tmp);

	if (strlen(tmp) < 9) return FALSE;

    if (strcmp(tmp + strlen(tmp) - 8, ".mp3.del" ) == 0)
        return TRUE;

    return FALSE;
}

bool isRecycledRec(char *name)
{
    strcpy(tmp, name);
    strlwr(tmp);

	if (strlen(tmp) < 9) return FALSE;

    if (strcmp(tmp + strlen(tmp) - 8, ".rec.del" ) == 0)
        return TRUE;

    return FALSE;
}

bool isRecycledThumbnail(char *name)
{
    strcpy(tmp, name);
    strlwr(tmp);

	if (strlen(tmp) < 8) return FALSE;

    if (strcmp(tmp + strlen(tmp) - 7, ".del.tn" ) == 0)
        return TRUE;

    return FALSE;
}

bool isCache(char *name)
{
    strcpy(tmp, name);
    strlwr(tmp);

	if (strlen(tmp) < 7) return FALSE;

    if (strcmp(tmp + strlen(tmp) - 6, ".cache" ) == 0)
        return TRUE;

    return FALSE;
}

bool isThumbnail(char *name)
{
    strcpy(tmp, name);
    strlwr(tmp);

	if (strlen(tmp) < 4) return FALSE;

    if (strcmp(tmp + strlen(tmp) - 3, ".tn" ) == 0)
        return TRUE;

    return FALSE;
}

bool isPhotoPlaylist(char *name)
{
    strcpy(tmp, name);
    strlwr(tmp);

	if (strlen(tmp) < 5) return FALSE;

    if (strcmp(tmp + strlen(tmp) - 4, ".ppl" ) == 0)
        return TRUE;

    return FALSE;
}

bool isMp3Playlist(char *name)
{
    strcpy(tmp, name);
    strlwr(tmp);

	if (strlen(tmp) < 5) return FALSE;

    if (strcmp(tmp + strlen(tmp) - 4, ".mpl" ) == 0)
        return TRUE;

    return FALSE;
}

bool isPlaylist(char *name)
{
    strcpy(tmp, name);
    strlwr(tmp);

	if (strlen(tmp) < 5) return FALSE;

    if (strcmp(tmp, "photo.pl") == 0 
			|| strcmp(tmp, "mp3.pl") == 0 
			|| strcmp(tmp + strlen(tmp) - 4, ".mpl" ) == 0
			|| strcmp(tmp + strlen(tmp) - 4, ".ppl" ) == 0)
        return TRUE;

    return FALSE;
}

bool isArtAssociation(char *name)
{
    strcpy(tmp, name);
    strlwr(tmp);

	if (strlen(tmp) < 5) return FALSE;

    if (strcmp(tmp + strlen(tmp) - 4, ".ass" ) == 0)
        return TRUE;

    return FALSE;
}

void ChangeDirRootFromUnknown()
{
	TYPE_File	fp;
	int		idx;
	extern int _appl_version;

	idx = 0;
	TAP_Hdd_FindFirst(&fp);

    //	loop until root found, allow maximum of 20 levels
	while ((strcmp(fp.name, "__ROOT__") != 0) && (idx++ < 20))
	{
		TAP_Hdd_ChangeDir("..");
		TAP_Hdd_FindFirst(&fp);
	}
	strcpy(tracked_dir, "/");
}

void ChangeToRoot(char *new_path)
{
	int	level, i;

	// if we are already at root, then nothing to do
	if (strcmp(tracked_dir, "/") == 0)
		return;

	level = getDirectoryLevel(tracked_dir);

	if (new_path != NULL) {
		if (strcmp(new_path, tracked_dir) == 0) {
			new_path[0] = '\0';
			return;
		}

		strcpy(log_buffer, tracked_dir);
		strcpy(tracked_dir, "/");
		do {
			ExtractDirectory(log_buffer, tmp3);
			if (strcmp(tmp3, "/") != 0 && strncmp(new_path, tmp3, strlen(tmp3)) == 0) {
				level -= getDirectoryLevel(tmp3);
				strcpy(buffer4, new_path);
				memset(new_path, '\0', 256);
				strncat(new_path, buffer4+strlen(tmp3), strlen(buffer4)-strlen(tmp3));
				strcpy(tracked_dir, tmp3);
				break;
			}
			strcpy(log_buffer, tmp3);
		}
		while (strcmp(tmp3, "/") != 0);
	}
	else {
		strcpy(tracked_dir, "/");
	}

    // change up until root reached
	i = level;
	while (level > 0)
	{
		//if (write_log) logText("Change dir ..", FALSE);
		TAP_Hdd_ChangeDir("..");

		level--;
	}
}

// convert all slashes in the path from '/' to '\', apart from the first one, which needs removing
void transformPathDelimiters(char *path, char *extract)
{
	//char result[100];
	size_t slash_idx;

	// return the text Root if are at '/'
	if (strcmp(path, "/") == 0) {
		strcpy(extract, "Root");
		return;
	}

	if (strncmp(path, "/", 1) == 0)
		strcpy(extract, path+1);
	else 
		strcpy(extract, path);

	slash_idx = strcspn(extract, "/");
	while (slash_idx > 0 && slash_idx < strlen(extract)) {
		extract[slash_idx] = '\\';
		slash_idx = strcspn(extract, "/");
	}

	//return result;
}

int containsSubstring(char *string_to_check, char *search_string)
{
	word i;

	if (strlen(search_string) == 0 || (strlen(search_string) > strlen(string_to_check))) return -1;

	strcpy(tmp, string_to_check);
	strcpy(tmp2, search_string);
    strlwr(tmp);
	strlwr(tmp2);

	i = 0; 
	while (i < strlen(tmp)-strlen(tmp2)+1) {
		if (strncmp(tmp+i, tmp2, strlen(tmp2)) == 0) {
			return i;
		}

		i++;
	}

	return -1;
}

// essentially extract the folder or file name from the full path
void ExtractLastField(char *path, char *significant_part)
{
	char *temp;
	memset(significant_part, 0, 100);

	temp = (char *) strrchr(path, '/');
	strcpy(significant_part, temp);
	if (strlen(significant_part) == 1)
		strcpy(significant_part, "/Root");
}

// essentially extract the extension of this file
void ExtractExtension(char *name, char *extension)
{
	char *temp;
	memset(extension, 0, 10);

	temp = (char *) strrchr(name, '.');

	if (temp == NULL) {
		strcpy(extension, "");
		return;
	}

	strcpy(extension, temp);
}

// essentially extract the extension of this file
int getExtensionLength(char *name)
{
	char extension[10];

	ExtractExtension(name, extension);

	return strlen(extension);
}

// extract the recording name from a scene name e.g. {rec_name}.rec.tn.2
void extractRecordingName(char *full_name, char *recording_name)
{
	char *temp;
	memset(recording_name, '\0', 100);

	temp = (char *) strrchr(full_name, 'c');
	strncat(recording_name, full_name, strlen(full_name)-strlen(temp)+1);
}

// extract the recording name from a scene name e.g. {rec_name}.rec.tn.2
void ExtractBaseMp3Directory(char *directory, char *base_dir)
{
	char *temp;
	memset(base_dir, '\0', 100);

	// if not a sub dir of MP3 then return any old rubbish to ensure no match is found
	if (strlen(directory) < 6 || strncmp(directory, "/MP3/", 5) != 0) {
		strcpy(base_dir, "XYZYX");
		return;
	}

	strcpy(tmp, directory+5);
	temp = (char *) strchr(tmp, '/');
	if (temp == NULL)
		strcpy(base_dir, tmp);
	else
		strncat(base_dir, tmp, strlen(tmp)-strlen(temp));
}


// essentially extract the directory from the full path including file name
void ExtractDirectory(char *path, char *significant_part)
{
	char *temp;
	memset(significant_part, 0, 256);

	temp = (char *) strrchr(path, '/');
	strncat(significant_part, path, strlen(path)-strlen(temp));
	if (strlen(significant_part) == 0) strcpy(significant_part, "/");
}

// truncate the string so that it is less than or equal to the available width
void truncateStringForDisplay(char *text, dword max_length, byte font_size)
{
	int reduced_length;
	strcpy(tmp3, text);

	while (getWidth(tmp3, font_size) > max_length) {
		reduced_length = strlen(tmp3)-1;
		memset(tmp3, '\0', 256);
		strncat(tmp3, text, reduced_length);
	}

	strcpy(text, tmp3);
}

bool ChangeDirPath(char *path)
{
	char *startPos;
	char *endPos;
	bool ready;

	//if (write_log) TAP_SPrint(log_buffer, "Change dir to %s from %s", path, tracked_dir);
	//if (write_log) logText(log_buffer, FALSE);

	strcpy(buffer5, path);
	ChangeToRoot(buffer5);

	startPos=buffer5;
	if ((*startPos)!='/')
		return FALSE;
	startPos=startPos+1;
	ready=FALSE;
	while (ready==FALSE){
		endPos=startPos;
		while (((*endPos)!=0) && ((*endPos)!='/')){
			endPos=endPos+1;
		}
		if ((*endPos)==0){
			ready=TRUE;
			if (!ChangeDir(startPos))
				return FALSE;
		} else {
			(*endPos)=0;
			if (!ChangeDir(startPos))
				return FALSE;
			(*endPos)='/';
			startPos=endPos+1;
		}
	}

	return TRUE;
}	

bool ChangeDirMp3(){
	return ChangeDirPath("/MP3");
}

bool ChangeDirProgramFiles(){
	return ChangeDirPath("/ProgramFiles");
}

bool ChangeDirSettings(){
#ifndef WIN32
	ChangeDirProgramFiles();
	if (!TAP_Hdd_Exist("Settings"))
		TAP_Hdd_Create("Settings", ATTR_FOLDER);
	return ChangeDir("Settings");
#else
	return ChangeDirPath("/ProgramFiles/Settings");
#endif
}

bool ChangeDirMediaManager(){
#ifndef WIN32
	ChangeDirSettings();
	if (!TAP_Hdd_Exist("MediaManager"))
		TAP_Hdd_Create("MediaManager", ATTR_FOLDER);
	return ChangeDir("MediaManager");
#else
	return ChangeDirPath("/ProgramFiles/Settings/MediaManager");
#endif
}

bool ChangeDirPlaylists(){
#ifndef WIN32
	ChangeDirMediaManager();
	if (!TAP_Hdd_Exist("Playlists"))
		TAP_Hdd_Create("Playlists", ATTR_FOLDER);
	return ChangeDir("Playlists");
#else
	return ChangeDirPath("/ProgramFiles/Settings/MediaManager/Playlists");
#endif
}

bool ChangeDirSkins(){
#ifndef WIN32
	ChangeDirSettings();
	if (!TAP_Hdd_Exist("Skins"))
		TAP_Hdd_Create("Skins", ATTR_FOLDER);
	return ChangeDir("Skins");
#else
	return ChangeDirPath("/ProgramFiles/Settings/Skins");
#endif
}

bool ChangeDirFonts(){
#ifndef WIN32
	ChangeDirSettings();
	return ChangeDir("Fonts");
#else
	return ChangeDirPath("/ProgramFiles/Settings/Fonts");
#endif
}

bool ChangeDirLogos(){
#ifndef WIN32
	ChangeDirSettings();
	if (!TAP_Hdd_Exist("Logos"))
		TAP_Hdd_Create("Logos", ATTR_FOLDER);
	return ChangeDir("Logos");
#else
	return ChangeDirPath("/ProgramFiles/Settings/Logos");
#endif
}

bool ChangeDirDataFiles(){
	return ChangeDirPath("/DataFiles");
}

bool ChangeDirCurrentRecordingRecursively(char *name, dword start_cluster, TYPE_File *rec_file)
{
	TYPE_File file;
	char **subfolders = dlmalloc(sizeof(*subfolders) * 128);
	int subfolder_count = 0, i;

	i = TAP_Hdd_FindFirst(&file);
	while (i--) {
		
		if (strcmp(file.name, name) == 0 && file.startCluster == start_cluster) {
			memcpy(rec_file, &file, sizeof(file));
			return TRUE;
		}

		if (file.attr == ATTR_FOLDER) {
			subfolders[subfolder_count] = dlmalloc(100 * sizeof(char));
			strcpy(subfolders[subfolder_count], file.name);
			subfolder_count++;
		}

		if (i == 0 || TAP_Hdd_FindNext(&file) == 0)
			break;
	}

	for (i = 0; i < subfolder_count; i++)
	{
		ChangeDir(subfolders[i]);
		if (ChangeDirCurrentRecordingRecursively(name, start_cluster, rec_file)) {
			for (i = 0; i < subfolder_count; i++)
			{
				dlfree(subfolders[i]);
				subfolders[i] = NULL;
			}
			return TRUE;
		}
		ChangeDir("..");
	}

	for (i = 0; i < subfolder_count; i++)
	{
		dlfree(subfolders[i]);
		subfolders[i] = NULL;
	}
	dlfree(subfolders);
	subfolders = NULL;

	dlmalloc_trim(0);
	return FALSE;
}

bool ChangeDirCurrentRecording(char *name, dword start_cluster, TYPE_File *rec_file)
{
	ChangeDirDataFiles();

	return ChangeDirCurrentRecordingRecursively(name, start_cluster, rec_file);
}

int getDirectoryLevel(char *directory)
{
	word i, levels = 0;

	for (i = 0; i < strlen(directory); i++) {
		if (directory[i] == '/')
			levels++;
	}

	return levels;
}

word countWordsInLine(char *info_line)
{
	int i = 0, word_count = 1;

	while (i <= strlen(info_line)) {
		if (info_line[i] == ' ')
			word_count++;
		i++;
	}

	return word_count;
}

void getWordInLine(char *info_line, char *word_in_line, word *line_pointer)
{
	int i = 0;
	memset(word_in_line, '\0', 40);

	while (i <= strlen(info_line) && i < 40-1) {
		word_in_line[i] = info_line[*line_pointer];
		if (i > 0 && word_in_line[i] == ' ') {
			*line_pointer = *line_pointer + 1;
			break;
		}

		*line_pointer = *line_pointer + 1;
		i++;
	}
}

bool getInfoLine(char *info_line, word *info_pointer, char *extInfo, int max_length, byte font_size)
{
	int i = 0;
	word last_space_pointer = 0, saved_info_pointer = *info_pointer;
	memset(info_line, '\0', 80);

	// cycle until at max width, end of text
	while (getWidth(info_line, font_size) < max_length 
			&& *info_pointer <= strlen(extInfo)
			&& i < 80-1) {

		// if we hit a line feed character, then ignore all these and flag so we can insert a blank line
		if (extInfo[*info_pointer] == 0x0a || extInfo[*info_pointer] == '\n') {
			*info_pointer = *info_pointer + 1;
			return TRUE;
		}

		// set char into the line
		info_line[i] = extInfo[*info_pointer];

		// if it is the end of a word, then save this pointer
		if (info_line[i] == ' ') last_space_pointer = i;

		// increment counters
		*info_pointer = *info_pointer + 1;
		i++;
	}

	// put pointer back to end of last word, and clear any text after that from the line
	if (*info_pointer < strlen(extInfo) && last_space_pointer > 0) {
		*info_pointer = *info_pointer - (*info_pointer-saved_info_pointer-last_space_pointer) + 1;
		memset(info_line+last_space_pointer, '\0', 80-last_space_pointer);
	}

	return FALSE;
}

bool isWokenFromTimer()
{
	return (*(word*)0xa3fffffc == 2);
}

/*
406 TF5000/5500
416 TF5000t
436 TF5010/TF5510
446 TF5000 Black Panther/White Polar
456 TF5800t MHEG
466 TF5000t Black Panther/White Polar
486 TF5010-SE
501 TF5000CI (ME)
1416 TF5000 Masterpiece
1426 TF5000t Masterpiece
1456 TF5010 Masterpiece
1501 TF5000CI (EUR)
10446 TF5200c
12406 TF5100c
13406 TF5100
1466 TF5100t_MP
1486 TF5100c_MP
*/

// Return the Topfield model number
bool isTF5800()
{
#ifdef WIN32
	return TRUE;
#endif               

    switch (*sysID)
    {
		case 456:	
		case 457:	
		case 458:	
			return TRUE;

		default:	
			return FALSE;
    }

    return FALSE;
}

bool isTF5100t()
{
#ifdef WIN32
	return FALSE;
#endif               

    switch (*sysID)
    {
		case 13406:  // TF5100
		case 1466:   // TF5100t_MP
		case 10426:  // PC5101t_5102t
			return TRUE;
		default:	
			return FALSE;
    }

    return FALSE;
}

bool isTF5100c()
{
#ifdef WIN32
	return FALSE;
#endif               

    switch (*sysID)
    {
		case 12406:	 // TF5100c
		case 1486:   // TF5100c_MP
		case 10416:  // PC5101c_5102c
			return TRUE;
		default:	
			return FALSE;
    }

    return FALSE;
}

bool isTF5200c()
{
#ifdef WIN32
	return FALSE;
#endif               

    switch (*sysID)
    {
		case 10446:  // TF5200c ?
			return TRUE;
		default:	
			return FALSE;
    }

    return FALSE;
}

bool isTF5000()
{
	return !isTF5800() && !isTF5100t() && !isTF5100c();
}

// Define character values
//  32=" "  35="#"  45="-"   46="."  47="/"  48="0"  57="9"  95="_"
#define CHAR_SPACE   32
#define CHAR_HASH    35
#define CHAR_ASTERIX 42
#define CHAR_PLUS    43
#define CHAR_DASH    45
#define CHAR_DOT     46
#define CHAR_SLASH   47
#define CHAR_ZERO    48
#define CHAR_NINE    57
#define CHAR_SCORE   95

bool MatchFileStamp(char* stamp, char* str)
{
     int d, l, start, stampLen, fileLen;
     char weekDay[12];
     
     stampLen = strlen( stamp );
     fileLen  = strlen( str ); // Length of filename.
     start    = fileLen - stampLen; // starting position to check in filename.
         
     // Handle the " DDD" date stamp.
     /*if ( strcmp( stamp, " DDD" ) == 0 )
     {
        for ( d = 0; d < 7; d++ )               
        {
            WeekdayToAlpha(d, weekDay);
            if ( strncmp( str + start + 1, weekDay, 3 ) == 0 )
                return TRUE;
        }
        return FALSE;
     }
     else
     // Handle the "- DDD" date stamp.
     if ( strcmp( stamp, "- DDD" ) == 0 )
     {
        for ( d = 0; d < 7; d++ )               
        {
            WeekdayToAlpha(d, weekDay);
            if ( strncmp( str + start + 2, weekDay, 3 ) == 0 )
                return TRUE;
        }
        return FALSE;
     }
     else
     {*/          
        // Handle all other date stamps.          
        for (l=0; l<stampLen; l++)
        {
            if ( stamp[l] == CHAR_ASTERIX )  // * is a wildcard for -_./ space
            {   
               if ( (str[start+l] != CHAR_DASH ) && (str[start+l] != CHAR_SCORE ) && ( str[start+l] != CHAR_DOT) && ( str[start+l] != CHAR_SLASH)&& ( str[start+l] != CHAR_SPACE))
                  return FALSE;
            }
            else 
            if ( stamp[l] == CHAR_SPACE ) 
            {   
               if ( (str[start+l] != CHAR_SPACE ) && (str[start+l] != CHAR_SCORE ))
                  return FALSE;
            }
            else 
            if ( stamp[l] == CHAR_PLUS ) 
            {   
               if ( str[start+l] != CHAR_PLUS )
                  return FALSE;
            }
            else 
            if ( stamp[l] == CHAR_HASH ) 
            {
                 if ( str[start+l] != CHAR_HASH)
                    return FALSE;
            }
            else 
            if ( stamp[l] == CHAR_DASH ) 
            {
                 if ( str[start+l] != CHAR_DASH )
                    return FALSE;
            }
            else 
            if ( stamp[l] == CHAR_DOT ) 
            {
                 if ( str[start+l] != CHAR_DOT )
                    return FALSE;
            }
            else 
            if ( stamp[l] == CHAR_SLASH ) 
            {
                 if ( str[start+l] != CHAR_SLASH )
                    return FALSE;
            }
            else 
            if ( stamp[l] == CHAR_SCORE ) 
            {
                 if ( str[start+l] != CHAR_SCORE )
                    return FALSE;
            }
            else 
            if ( stamp[l] == CHAR_NINE ) 
            {
                 if (( str[start+l] < CHAR_ZERO ) || ( str[start+l] > CHAR_NINE))
                    return FALSE;
            }
            else 
            if ((str[start+l] < CHAR_ZERO) || (str[start+l] > CHAR_NINE)) 
               return FALSE;
        } 
     //}  
         
     return TRUE;            
}

bool hasDateStamp(char* name)
{
	char stamps[15][20] = {"*99*99 YYYY*MM*DD", "*YYYYMMDD_9999", "*YYYYMMDD", "*YYYY*MM*DD", "*YY*MM*DD", "*DD*MM*YYYY", "*DD*MM*YYYY*9999", "*DD*MM*YYYY*99*99", "*DDMMYY", "*DDMMYY9999", "*DDMMYY*9999", "*DD*MM", "*MMDD", "- DDD", " DDD"};
    word i;
          
    for (i = 0; i < 12; i++)  // Go through each date/time stamp and see if we have a matching filename
    {
        if (MatchFileStamp(stamps[i], name)) // We've found a match.
        {
            return TRUE;
        } 
	}

	return FALSE;
}

int isCopy(char *name)
{
	int idx = containsSubstring(name, "(COPY)");

	if (idx >= 0 && (((int) strlen(name) - 8 - idx) >= 0))
		return idx;
	
	return -1;
}

void FormatSortName(char* target, char* source)
{
	#define NUMBER_OF_STAMPS 21   
  
	/*
	* = -_./ space
	9 = 0-9

	*99*99 YYYY*MM*DD
	*YYYYMMDD_9999
	*YYYYMMDD
	YYYYMMDD
	*YYYY*MM*DD
	YYYY*MM*DD
	*YY*MM*DD (also matchs *DD*MM*YY )
	YY*MM*DD
	*DD*MM*YYYY
	*DD*MM*YYYY*9999
	*DDMMYY
	DDMMYY
	*DD*MM
	DD*MM
	- DDD
	*DDD
	DDD

	-99
	_99
	-9
	_9

	_+
	-+
	_#
	-#
	*/

	strcpy(target, source);

	// check for a single digit start, in which case add a leading zero
	if (strlen(source) > 1) {
		if ((source[0] >= 48 && source[0] <= 57) && (source[1] < 48 || source[1] > 57)) {
			memset(target, '\0', strlen(target));
			strcpy(target, "0");
			strcat(target, source);
		}
	}

	if (isRec(target)) {
		char stamps[NUMBER_OF_STAMPS][20] = {"_#", "-#", "_+", "-+", "*99*99 YYYY*MM*DD", "*YYYYMMDD_9999", "*YYYYMMDD", "*YYYY*MM*DD",  "*YY*MM*DD", "*DD*MM*YYYY", "*DD*MM*YYYY*9999", "*DD*MM*YYYY*99*99",  "*DDMMYY", "*DDMMYY9999", "*DDMMYY*9999", "*DD*MM",  "*MMDD",  "- DDD", " DDD", "-99", "-9"};
		int i, start, stampLen, fileLen;

		//strcpy( target, source);                   // Copy the whole filename by default.
		fileLen = strlen(target);                  // Get the length of the filename.
		target[fileLen - 4]='\0';                  // Chop off the ".rec" from the filename.
	
		// strip out the (COPY)-x extension for a decoded file
		start = isCopy(target);
		if (start >= 0) {
			target[start]='\0';  
			memcpy(tmp, target+start+8, fileLen-4-start-8);
			tmp[fileLen-4-start-8]='\0'; 
			strcat(target, tmp);
		}

		for (i=0; i < NUMBER_OF_STAMPS; i++)  // Go through each date/time stamp and see if we have a matching filename
		{
			if (MatchFileStamp(stamps[i], target)) // We've found a match.
			{
				stampLen = strlen( stamps[i] );     // Get the length of the date/time stamp.
				fileLen  = strlen( target );        // Get the length of filename.
				start    = fileLen - stampLen;      // Work out where the real file name ends and date/time stamp begins.
				target[start]='\0';                 // Terminate the filename at the date/time stamp beginning.
				break;
			} 
		}
	}
}

bool OsdActive (int startCol, int startRow, int endCol, int endRow)
{
#ifndef WIN32
	TYPE_OsdBaseInfo osdBaseInfo;
    dword* wScrn;
    dword iRow, iCol;

    TAP_Osd_GetBaseInfo( &osdBaseInfo );

    for ( iRow = startRow; iRow < endRow; iRow += 4 )  //every 4th line
    {
        wScrn = osdBaseInfo.eAddr + 720 * iRow;

        for ( iCol = startCol; iCol < endCol; iCol += 6 ) //may as well only scan every 6 pixels to save time and resource, stop at 350 as new f/w puts channel details to the right.
        {
            if ( (*(wScrn + iCol)) != 0 )
                return TRUE; // Do not allow Key Action
        }
    }
#endif

    return FALSE;
}

// convert utc to local time, either using gmt offset from flash, or hack if this is not present
void gmtToLocalTime(dword utc_time, word *local_mjd, word *local_hour, word *local_min, int offset, dword offset_override)
{
	int offset_to_use = offset_override > 0 ? offset_override : offset;
	word year, mjd; byte month, day, week_day, hour, min, sec, lsM, lsO; bool its_summer = FALSE;

	if (offset_to_use > 0 && offset_to_use < 24) {
		*local_mjd	= (utc_time >> 16) & 0xffff;
        *local_hour	= (utc_time >> 8) & 0xff;  
		*local_min	= utc_time & 0xff;  

		TAP_GetTime(&mjd, &hour, &min, &sec);
		TAP_ExtractMjd(mjd, &year, &month, &day, &week_day);

		lsM = (byte) (31 - ((int) (5 * year / 4) + 4) % 7);            
		lsO = (byte) (31 - ((int) (5 * year / 4) + 1) % 7);

		if ((month > 3 && month < 10) || (month == 3 && day >= lsM) || (month == 10 && day < lsO))
			its_summer = TRUE;

		if (offset_override == 0) {
			TAP_ExtractMjd(*local_mjd, &year, &month, &day, &week_day);

			lsM = (byte) (31 - ((int) (5 * year / 4) + 4) % 7);            
			lsO = (byte) (31 - ((int) (5 * year / 4) + 1) % 7);

			if ((month > 3 && month < 10) || (month == 3 && day >= lsM) || (month == 10 && day < lsO)) {
				if (!its_summer)
					offset_to_use = offset_to_use + 1;
			}
			else {
				if (its_summer)
					offset_to_use = offset_to_use - 1;
			}
		}

		if (*local_hour + offset_to_use >= 24) {
			*local_hour	= offset_to_use - (24 - *local_hour);
			*local_mjd	= *local_mjd + 1;
		}
		ef (*local_hour + offset_to_use < 0) {
			*local_hour	= 24 + (*local_hour + offset_to_use);
		}
		else
			*local_hour	= *local_hour + offset_to_use;
	}
	else {
		*local_mjd	= (utc_time >> 16) & 0xffff;
        *local_hour	= (utc_time >> 8) & 0xff;  
		*local_min	= utc_time & 0xff;  

		TAP_ExtractMjd(*local_mjd, &year, &month, &day, &week_day);

		lsM = (byte) (31 - ((int) (5 * year / 4) + 4) % 7);            
		lsO = (byte) (31 - ((int) (5 * year / 4) + 1) % 7);

		if ((month > 3 && month < 10) || (month == 3 && day >= lsM) || (month == 10 && day < lsO)) {
			if (*local_hour + 1 >= 24) {
				*local_hour	= 1 - (24 - *local_hour);
				*local_mjd	= *local_mjd + 1;
			}
			else
				*local_hour	= *local_hour + 1;
		}
	}
}

/*void adjustEndTime(word start_hour, word start_min, word *end_hour, word *end_min, word duration)
{
	word start_in_mins, end_in_mins;
	
	// convert start and end time to mins
	start_in_mins = start_hour*60 + start_min;
	end_in_mins = *end_hour*60 + *end_min;

	// add 24 hours to end if it is the next day
	if (end_in_mins < start_in_mins) 
		end_in_mins += 24*60;
	
	// if end is within 45 mins of start+duration, then assume it is ok
	if ((end_in_mins - start_in_mins) < (duration + 45))
		return;

	// recalculate end as (start + duration)
	end_in_mins = start_in_mins + duration;
	end_in_mins = end_in_mins % (24*60);

	*end_hour = end_in_mins / 60;
	*end_min = end_in_mins % 60;
}*/

#ifndef WIN32
/*void DeterminePIN(char *pin, dword eeprom_offset)
{
	word       SysID = *((volatile word*)0xa3fffffa);
	extern int _appl_version;
	dword      eEPROMOffset=0;
	word       nPin;

	// old method relied on hard coded offsets per model/firmware
	TAP_Print("SysID = %d, _appl_version = %d\r\n", SysID, _appl_version);
	if ((SysID == 1416) && (_appl_version == 0x1203)) eEPROMOffset=0x81b0ee2c;  // (5000 FW 5.12.03)
	if ((SysID == 456)  && (_appl_version == 0x1204)) eEPROMOffset=0x81bc3c8c;
	if ((SysID == 456)  && (_appl_version == 0x1205)) eEPROMOffset=0x81bc3e0c;
	if ((SysID == 456)  && (_appl_version == 0x1209)) eEPROMOffset=0x81c0e76c;
	if ((SysID == 456)  && (_appl_version == 0x1225)) eEPROMOffset=0x81c127d4;
	if ((SysID == 456)  && (_appl_version == 0x1288)) eEPROMOffset=0x81daa2c4;
 
	if (eeprom_offset == 0)	{
		TAP_Print ("Unsupported firmware version.\r\n");
		pin[0]='\0'; 
		return;
	}

	// Now simply read the pin from the EEPROM into a variable
	//nPin =  *((word*) (eEPROMOffset)+0x12);
	//TAP_SPrint (pin,"%4.4d", nPin);

	// new method looks up the offset on startup
	nPin =  *((word*) (eeprom_offset)+0x12);
	TAP_SPrint (pin,"%4.4d", nPin);
}*/
#endif

/*dword getEEPROMOffset()
{
	dword   gph = 0, gpl = 0, gp, *p;
	dword   pEEPROM = 0;
 
	#define Cmd_JR_RA 0x03e00008
	#define Mask_LW_GP 0xffe00000
	#define Cmd_LW_GP 0x8f800000
	#define Cmd_LUI_GP 0x3c1c0000
	#define Cmd_ADDIU_GP 0x279c0000
 
	//Get the firmwares $gp
	p = (dword *) 0x80000000;
	do
	{
		if ((*p & 0xffff0000) == Cmd_LUI_GP) gph = *p & 0xffff;
		if ((*p & 0xffff0000) == Cmd_ADDIU_GP) gpl = *p & 0xffff;
		p++;
	} while ((gpl == 0) || (gph == 0));
	gp = (gph << 16) + (short) gpl;
 
	//Scan for the "lw $??, ????($gp)" instruction in TAP_Channel_GetCurrent
	p = (dword *) TAP_Channel_GetCurrent;
	do
	{
		if ((*p & Mask_LW_GP) == Cmd_LW_GP)
		{
			pEEPROM = gp + (short) (*p & 0x0000ffff);
			break;
		}
		p++;
	} while (*p != Cmd_JR_RA);
	if (pEEPROM != 0) pEEPROM = *(dword *) pEEPROM - 2;

	return pEEPROM;
}*/


